from datetime import datetime
from cy_components.helpers.formatter import DateFormatter
from cy_components.utils.notifier import MessageHandler, MessageType
from cy_widgets.logger.base import RecorderBase
from cy_data_access.models.log import *


class ProcedureRecorder(RecorderBase):

    def __init__(self, m_token: str, m_type: MessageType):
        self.__message_token = m_token
        self.__message_type = m_type

    def record_simple_info(self, content):
        print(content, end='\n\n')

    def record_procedure(self, content):
        content = DateFormatter.now_date_string('[%H:%M:%S] ') + content
        print(content, end='\n\n')

    def record_exception(self, content):
        print(content, end='\n\n')
        MessageHandler.send_message(content, 'Precedure Exception', self.__message_type, self.__message_token)

    def _record_summary_log(self):
        MessageHandler.send_message(self._summary_log, 'Procedure Summary', self.__message_type, self.__message_token)


class PersistenceRecorder(RecorderBase):

    def __init__(self, m_token: str, m_type: MessageType, log_type: LogType):
        self.__message_token = m_token
        self.__message_type = m_type
        self.__log_type = log_type

    def record_simple_info(self, content):
        print(content, end='\n\n')

    def record_procedure(self, content):
        content = DateFormatter.now_date_string('[%H:%M:%S] ') + content
        print(content, end='\n\n')

    def record_exception(self, content):
        print(content, end='\n\n')
        MessageHandler.send_message(content, 'Precedure Exception', self.__message_type, self.__message_token)

    def _record_summary_log(self):
        info = LogInfo()
        info.log = self._summary_log
        info.log_type = self.__log_type
        info.create = datetime.now()
        try:
            info.save()
        except Exception:
            self.record_exception("Log save failed")



from enum import IntEnum
from cy_components.helpers.formatter import DateFormatter
from cy_widgets.fetcher.exchange import *


class ExchangeFetchingType(IntEnum):
    HISTORICAL = 0,     # Historical candle data, from earliest date do backward fetching
    FILL_RECENTLY = 1,  # Fill data recently
    CHECK_MISSING = 2,  # NOT IMPLEMENTED. Check missing candle
    TRADING = 3,        # NOT IMPLEMENTED. For trading


class ExchangeFetchingConfiguration:
    """ 抓取过程需要的配置参数
    1. CoinPair;
    2. Operation Type;
    3. TimeFrame;
    4. Begin date;
    5. Fetching duration (Sleep after a fetching operation)
    6. fetching_by_ccxt
    """

    def __init__(self,
                 coin_pair: CoinPair = None,
                 time_frame=TimeFrame.Minute_5,
                 sleep_duration=5,
                 op_type=ExchangeFetchingType.HISTORICAL,
                 fetching_by_ccxt=True,
                 batch_limit=1000,
                 debug=False):
        super().__init__()
        assert coin_pair is not None
        assert time_frame is not None
        assert sleep_duration > 0

        self.coin_pair = coin_pair
        self.time_frame = time_frame
        self.op_type = op_type
        self.sleep_duration = sleep_duration
        self.fetching_by_ccxt = fetching_by_ccxt
        self.batch_limit = batch_limit if batch_limit > 20 else 50
        self.debug = debug


class ExchangeFetchingProcedure:
    """抓取数据的一般流程"""

    def __init__(self,
                 fetcher: ExchangeFetcher,
                 configuration: ExchangeFetchingConfiguration,
                 get_earliest_date,
                 get_latest_date,
                 save_df):
        """初始化

        Parameters
        ----------
        fetcher : ExchangeFetcher
            抓取K线对象
        configuration : ExchangeFetchingConfiguration
            抓取配置
        get_earliest_date : () -> Date
            获取最早K线的日期
        get_latest_date : () -> Date
            获取最近K线的日期
        save_df : (df) -> Bool
            保存K线数据，返回是否继续抓取
        """
        assert fetcher is not None
        assert configuration is not None
        assert save_df is not None

        self.fetcher = fetcher
        self.configuration = configuration

        self.get_earliest_date = get_earliest_date
        self.get_latest_date = get_latest_date
        self.save_df = save_df

    # Data

    def __perform_fetching(self, since_ts):
        """fetching + saving"""
        # Fetch
        df = self.fetcher.fetch_historical_candle_data(
            self.configuration.coin_pair,
            self.configuration.time_frame,
            since_ts,
            self.configuration.batch_limit,
            by_ccxt=self.configuration.fetching_by_ccxt)
        return self.save_df(df)

    # Task

    def __fetch_historical_data(self):
        """获取历史记录"""
        while True:
            assert self.get_earliest_date is not None
            # 获取最早的日期
            earliest_date = self.get_earliest_date()
            # 转为时间戳
            earliest_ts = DateFormatter.convert_local_date_to_timestamp(earliest_date)
            # 确定抓取的起始时间戳
            since_ts = self.configuration.time_frame.timestamp_backward_offset(
                earliest_ts, self.configuration.batch_limit)
            # 需要继续的，暂停一会儿
            if self.__perform_fetching(since_ts):
                time.sleep(self.configuration.sleep_duration)
            else:
                print('Historical.finished')
                return

    def __fill_recently(self):
        """补齐最近的数据"""
        while True:
            assert self.get_latest_date is not None
            # 获取已经保存的最新的日期
            recent_date = self.get_latest_date()
            # 转换到时间戳
            recent_ts = DateFormatter.convert_local_date_to_timestamp(recent_date)
            # 往前移动10个单位作为起始日期
            since_ts = self.configuration.time_frame.timestamp_backward_offset(recent_ts, 10)
            # 需要继续的，暂停一会儿
            if self.__perform_fetching(since_ts):
                time.sleep(self.configuration.sleep_duration)
            else:
                print('FillRecently.finished')
                return

    def run_task(self):
        """Dispatch task"""
        if self.configuration.op_type == ExchangeFetchingType.HISTORICAL:
            self.__fetch_historical_data()
        elif self.configuration.op_type == ExchangeFetchingType.FILL_RECENTLY:
            self.__fill_recently()



from datetime import datetime
from cy_components.helpers.formatter import DateFormatter as dfr
from cy_components.defines.column_names import *
from cy_widgets.backtest.strategy import *
from cy_widgets.backtest.helper import *
from cy_data_access.models.backtest import *
from multiprocessing.pool import Pool


class SimpleBacktestProcedure:
    """简单回测流程"""

    def __init__(self, coin_pair, df, strategy_cls, params, position_func, evaluation_func):
        self.__df = df
        self.__strategy_cls = strategy_cls
        self.__params = params
        self.__position_func = position_func
        self.__evaluation_func = evaluation_func
        self.__collection_name = '{}_{}_{}'.format(dfr.convert_local_date_to_string(
            datetime.now(), '%Y%m%d%H%M%S'), coin_pair.formatted('_').lower(), str(strategy_cls()).lower())

        self.__start_date = df.iloc[0][COL_CANDLE_BEGIN_TIME]
        self.__end_date = df.iloc[-1][COL_CANDLE_BEGIN_TIME]

    def __result_handler(self, evaluated_df, strategy, error_des):
        if error_des is not None:
            print(error_des)
        else:
            param = StrategyHelper.formatted_identifier(strategy)
            curve = str(evaluated_df.iloc[-1][COL_EQUITY_CURVE])
            print('finished', param, curve)
            summary_collection_cls = backtest_summary_class(self.__collection_name)
            summary_collection_cls(
                parameter=param,
                equity_curve=curve,
                start_date=self.__start_date,
                end_date=self.__end_date
            ).save()

    def calculation(self, param):
        try:
            bt = StrategyBacktest(self.__df.copy(), self.__strategy_cls(
                **param), self.__position_func, self.__evaluation_func, self.__result_handler)
            return bt.perform_test()
        except Exception as e:
            print(str(param), str(e))
            return None

    def perform_test_proc(self, processes=2):
        # self.calculation(self.__params[0])
        with Pool(processes=processes) as pool:
            start_date = datetime.now()
            pool.map(self.calculation, self.__params)
            print(datetime.now() - start_date)



from .aip_base import *
from ..exchange.binance import *


class BinanceAIP(AIPBase):
    """币安定投"""

    def __init__(self,
                 coin_pair,
                 time_frame,
                 signal_provider,
                 day_of_week,
                 ma_periods,
                 trader_provider,
                 invest_base_amount,
                 recorder,
                 debug=False,
                 fee_percent=0.0015):
        super().__init__(coin_pair, time_frame, signal_provider, day_of_week, ma_periods,
                         trader_provider, invest_base_amount, recorder, debug=debug, fee_percent=0)
        self.__handler = BinanceHandler(trader_provider)

    def _prepare_to_buying(self, invest_amount):
        """下单前准备，失败自行记录:
        self.recorder.record_summary_log('**划转{}失败**'.format(self.coin_pair.base_coin.upper()))
        """
        try:
            self.__handler.transfer_margin(self.coin_pair.base_coin, invest_amount, 2)
        except Exception:
            self.recorder.record_summary_log('**划转{}失败**'.format(self.coin_pair.base_coin.upper()))

    def _place_buying_order(self, invest_amount):
        """下单，成功返回：
        {
        'price': 123,
        'cost': 123,
        'amount': 123
        }"""
        try:
            logger = TraderLogger(self.trader_provicer.display_name, self.coin_pair.formatted(), 'Spot', self.recorder)
            return self.__handler.handle_spot_buying(self.coin_pair, invest_amount, logger)
        except Exception as e:
            self.recorder.record_summary_log("下单失败，" + str(e))

    def _rollback_when_order_failed(self, invest_amount):
        """下单失败后回滚"""
        try:
            self.__handler.transfer_margin(self.coin_pair.base_coin, invest_amount, 1)
        except Exception:
            self.recorder.record_summary_log('**回滚划转{}失败**'.format(self.coin_pair.base_coin.upper()))

    def _finishing_aip(self, remaining_base_coin_amount, order_amount):
        """完成定投后的收尾工作"""
        try:
            self.__handler.transfer_margin(self.coin_pair.base_coin, remaining_base_coin_amount, 1)
            self.__handler.transfer_margin(self.coin_pair.trade_coin, order_amount, 1)
        except Exception:
            self.recorder.record_summary_log('**定投结束后划转失败**')



import math
from cy_widgets.exchange.provider import *
from cy_widgets.trader.exchange_trader import *


class BinanceHandler:

    def __init__(self, ccxt_provider: CCXTProvider):
        self.__ccxt_provider = ccxt_provider
        self.__lending_products = None
        self.__fee_percent = 0  # 订单里已经把手续费扣掉返回了，不需要计算

    def fetch_all_lending_product(self):
        """所有活期产品"""
        # https://binance-docs.github.io/apidocs/spot/cn/#baa37cb2f9
        # [{
        #    "asset": "BTC",
        #    "avgAnnualInterestRate": "0.00250025",
        #    "canPurchase": true,
        #    "canRedeem": true,
        #    "dailyInterestPerThousand": "0.00685000",
        #    "featured": true,
        #    "minPurchaseAmount": "0.01000000",
        #    "productId": "BTC001",
        #    "purchasedAmount": "16.32467016",
        #    "status": "PURCHASING",
        #    "upLimit": "200.00000000",
        #    "upLimitPerUser": "5.00000000"
        # }, {...}]
        self.__lending_products = self.__ccxt_provider.ccxt_object_for_query.sapi_get_lending_daily_product_list()

    def daily_lending_product(self, coin_name):
        """查找对应币种的活期产品"""
        if self.__lending_products is None:
            self.fetch_all_lending_product()
        return list(filter(lambda x: x['asset'].lower() == coin_name.lower(), self.__lending_products))

    def fetch_balance(self, type="spot"):
        """查询余额
        type: spot/margin/future"""
        return self.__ccxt_provider.ccxt_object_for_query.fetch_balance({'type': type})

    def transfer_margin(self, coin_name, amount, type=1):
        """全仓杠杆账户划转

        Parameters
        ----------
        type : int, optional
            1: in, 2: out
        """
        return self.__ccxt_provider.ccxt_object_for_query.sapi_post_margin_transfer({
            'asset': coin_name,
            'amount': amount,
            'type': type
        })

    def handle_spot_buying(self, coin_pair, amount, trader_logger):
        """现货买入"""
        order = Order(coin_pair, amount, 0)  # Only set base coin amount
        executor = ExchangeOrderExecutorFactory.executor(self.__ccxt_provider, order, trader_logger)
        # 下单
        response = executor.handle_long_order_request()
        if response is None:
            raise ConnectionError("Request buying order failed")
        # Binance 手续费已经扣掉了
        # {'id': '701037299',
        # 'clientOrderId': 'Se8IFlpHyWpsY7OaYkhKC1',
        # 'timestamp': 1597589153872,
        # 'datetime': '2020-08-16T14:45:53.872Z',
        # 'lastTradeTimestamp': None,
        # 'symbol': 'BNB/USDT',
        # 'type': 'limit',
        # 'side': 'buy',
        # 'price': 23.3236,
        # 'amount': 0.47,
        # 'cost': 10.854319,
        # 'average': 23.094295744680853,
        # 'filled': 0.47,
        # 'remaining': 0.0,
        # 'status': 'closed',
        # 'fee': None,
        # 'trades': None}
        price = response['average']
        cost = response['cost']
        filled = response['filled']
        buy_amount = math.floor(filled * (1 - self.__fee_percent) * 1e8) / 1e8  # *1e8 向下取整再 / 1e8
        return {
            'price': price,
            'cost': cost,
            'amount': buy_amount
        }

    def handle_spot_selling(self, coin_pair, amount, trader_logger):
        """现货卖出"""
        order = Order(coin_pair, 0, amount, side=OrderSide.SELL)  # Only set trade coin amount
        executor = ExchangeOrderExecutorFactory.executor(self.__ccxt_provider, order, trader_logger)
        # place order
        response = executor.handle_close_order_request()
        if response is None:
            raise ConnectionError("Request selling order failed")
        price = response['average']
        cost = response['cost']
        filled = response['filled']
        return {
            'price': price,
            'cost': cost,
            'amount': filled
        }
